/*
  Java Japanese-English dictionary/Kanjidic browser
  Copyright (C) 1997 Todd David Rudick

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import JavaDict;
import GenList;
import ButtonList;

class LookupAction implements ActionListener {
  String dict, dictindex;
  Label LookupLabel;
  JavaDict javadict;
  TextField Entry;
  GenList outputList;
  Choice InputModeChoice;

  public void setDict(String dict) {
    this.dict = dict;
  }
  
  public LookupAction(Choice InputModeChoice,
		      TextField Entry,
		      Label LookupLabel, JavaDict javadict,
		      GenList outputList,
		      String dict, String dictindex) {
    this.Entry = Entry;
    this.LookupLabel = LookupLabel;
    this.javadict = javadict;
    this.outputList = outputList;
    this.InputModeChoice=InputModeChoice;
    this.dict = dict;
    this.dictindex = dictindex;
  }

  public void actionPerformed(ActionEvent e) {
    if (RKDisplay.firsttime) return;
    String lookup;
    String entrytext;
    java.lang.Thread t;
    lookup = LookupLabel.getText();
    entrytext = Entry.getText();
    // Fix up trailing matches if in hiragana or katakana mode
    if (((InputModeChoice.getSelectedItem()==GUI.HModeLabel)||
	 (InputModeChoice.getSelectedItem()==GUI.KModeLabel))&&
	(entrytext.length()>0))
      {
	int mode;
	if (InputModeChoice.getSelectedItem()==GUI.HModeLabel) mode=0;
	else mode=1;
	
	String s = entrytext+'\'';
	String LastDisplay = lookup;
	String LastEntry = entrytext;
	GUI.RomKanaPair FoundOne = null;
	int i,j;
	j = GUI.Rom2KanaVector.size();
	GUI.RomKanaPair rk=null;
	for (i=0;i<j;i++) {
	  try {
	    rk = (GUI.RomKanaPair) GUI.Rom2KanaVector.elementAt(i);
	    if (rk.mode == mode) {
	      if (s.startsWith(rk.romaji)) FoundOne = rk;
	    }
	  } catch (Exception e3) { 
	    System.out.println(e3.getMessage()); System.exit(1); }
	}
	if (FoundOne != null) {
	  LookupLabel.setText(LastDisplay+FoundOne.kana);
	}
	lookup = LookupLabel.getText();
      }
    

    // reset entry form
    LookupLabel.setText("");
    Entry.setText("");
    if (lookup.trim().length()==0) return;
    /*    Vector filtVector = new Vector();
	  
	  if (javadict.gui.romajiFilterBox.getState()) {
	  Filter.romajiFilter romfilt = 
	  new Filter.romajiFilter(GUI.Rom2KanaVector);
	  filtVector.addElement(romfilt);
	  }*/

    t = new java.lang.Thread(JavaDict.LookUp.lookupgroup,
			     new JavaDict.LookUp(lookup,outputList,
						 dict, dictindex,true,
						 javadict));
			     
    t.setDaemon(true); // exit if it's the only thing running
    t.setPriority(java.lang.Thread.MIN_PRIORITY);
    
    t.start();
    //    javadict.lookup(lookup,outputList);
  }
}

class RKDisplay implements KeyListener {
  Choice ModeChoice;
  TextField Entry;
  Label Display;
  int mode;
  String LastEntry = "";
  String LastDisplay = "";
  public static boolean firsttime = true;

  public RKDisplay(Choice ModeChoice, TextField Entry, Label Display) {
    this.ModeChoice = ModeChoice;
    this.Entry = Entry;
    this.Display = Display;
  }

  public void keyPressed(KeyEvent e) {
    if (firsttime) {
      Display.setText("");
      firsttime=false;
    }
    flushEntry();
    // mode change
    if (e.getKeyCode()==KeyEvent.VK_F3) {
      int ic = ModeChoice.getItemCount();
      int si = ModeChoice.getSelectedIndex();
      si = si + 1;
      if (si==ic) si=0;
      ModeChoice.select(si);
      return;
    }
    
    LastDisplay = Display.getText();
    if ((e.getKeyCode()==KeyEvent.VK_BACK_SPACE)&&
	(LastEntry.length()==0)&&
	(LastDisplay.length()!=0)) {
      try {
	Display.setText(LastDisplay.substring
			(0,LastDisplay.length()-1));
      } catch (Exception e2) {
	Display.setText("");
      }
      LastEntry = Entry.getText();
    }
    LastEntry = Entry.getText();
    LastDisplay = Display.getText();
  }

  void flushEntry() {
    String disp;
    boolean finished=false;
    while (((disp=Entry.getText()).length()!=0)&&(!finished)) {
      char curChar = disp.charAt(0);
      LastDisplay = Display.getText();
      {
        LastEntry = Entry.getText();

        if (ModeChoice.getSelectedItem()==GUI.EModeLabel) {
          Display.setText(LastDisplay+LastEntry);
          Entry.setText("");
        } else {
          if (ModeChoice.getSelectedItem()==GUI.HModeLabel)
            mode = 0;
          else if (ModeChoice.getSelectedItem()==GUI.KModeLabel)
            mode = 1;
          else {
            System.out.println("Internal error: unknown input mode (" +
                               ModeChoice.getSelectedItem() + ")");
            System.exit(1);
          }
          // so now we know the mode
          while (LastEntry.length()>0) { // handle the conversion (if poss.)
            String s = LastEntry;
  
            if (s.length()>1) {
              char c2 = ' ';
              try {c2 = s.charAt(1);} catch (Exception e4) {
                System.out.println(e4.getMessage()); System.exit(1); }
                /* if there is an "nn", "nm" or "mm", convert to first to "q"
                 to force it to be
                 converted to a kana "n"   */
              if (s.startsWith("nn")||s.startsWith("nm")||
                  s.startsWith("mm")) s = 'q' + s.substring(1);
  
              /* if there is "nx", where "x" is not a vowel or a "y", 
                 force the "n".  */
              if (s.startsWith("n")||(s.startsWith("m"))) {
                switch (c2) {
                case 'a':
                case 'e':
                case 'i':
                case 'o':
                case 'u':
                case 'y':
                case '\'':
                  break;
                default:
                  s = 'q' + s.substring(1);
                }
              }
              /* double consonant - force small "tsu" */
              char c1 = ' ';
              try {c1 = s.charAt(0);} catch (Exception e5) {
                System.out.println(e5.getMessage()); System.exit(1); }
              if (c1==c2) {
                s = '*' + s.substring(1);
              }
            }
  
  
            GUI.RomKanaPair FoundOne = null;
            boolean NoneBetter = true;
  
            int i,j;
            j = GUI.Rom2KanaVector.size();
            GUI.RomKanaPair rk=null;
            for (i=0;i<j;i++) {
              try {
                rk = (GUI.RomKanaPair) GUI.Rom2KanaVector.elementAt(i);
                if (rk.mode == mode) {
                  if (s.startsWith(rk.romaji)) FoundOne = rk; else
                    if (rk.romaji.startsWith(s)) NoneBetter = false;
                } // else System.out.println(rk.mode + ", " + mode);
              } catch (Exception e3) { 
                System.out.println(e3.getMessage()); System.exit(1); }
            }
            if ((FoundOne != null) && (NoneBetter)) {
              Display.setText(LastDisplay+FoundOne.kana);
              Entry.setText(LastEntry.substring(FoundOne.romaji.length()));
              Entry.setCaretPosition(Entry.getText().length());
              LastEntry = Entry.getText();
              LastDisplay = Display.getText();
            } else {
              finished=true;
              break;
            } /* leave the loop when we can't do more */
          }
        }
      }
    LastEntry = Entry.getText();
    LastDisplay = Display.getText();
       

    } // while

  }

  public void keyTyped(KeyEvent e) {

    if (firsttime) {
      Display.setText("");
      firsttime=false;
    }

    char curChar = e.getKeyChar();

    LastDisplay = Display.getText();


    {
      LastEntry = Entry.getText();
//      System.out.println("Entry box contains: " + LastEntry);

      if (ModeChoice.getSelectedItem()==GUI.EModeLabel) {
	Display.setText(LastDisplay+LastEntry);
	Entry.setText("");
      } else {
	if (ModeChoice.getSelectedItem()==GUI.HModeLabel)
	  mode = 0;
	else if (ModeChoice.getSelectedItem()==GUI.KModeLabel)
	  mode = 1;
	else {
	  System.out.println("Internal error: unknown input mode (" +
			     ModeChoice.getSelectedItem() + ")");
	  System.exit(1);
	}
	// so now we know the mode
	while (LastEntry.length()>0) { // handle the conversion (if poss.)
	  String s = LastEntry;

	  if (s.length()>1) {
	    char c2 = ' ';
	    try {c2 = s.charAt(1);} catch (Exception e4) {
	      System.out.println(e4.getMessage()); System.exit(1); }
	    /* if there is an "nn", "nm" or "mm", convert to first to "q"
	       to force it to be
	       converted to a kana "n"   */
	    if (s.startsWith("nn")||s.startsWith("nm")||
		s.startsWith("mm")) s = 'q' + s.substring(1);

	    /* if there is "nx", where "x" is not a vowel or a "y", 
	       force the "n".  */
	    if (s.startsWith("n")||(s.startsWith("m"))) {
	      switch (c2) {
	      case 'a':
	      case 'e':
	      case 'i':
	      case 'o':
	      case 'u':
	      case 'y':
	      case '\'':
		break;
	      default:
		s = 'q' + s.substring(1);
	      }
	    }
	    /* double consonant - force small "tsu" */
	    char c1 = ' ';
	    try {c1 = s.charAt(0);} catch (Exception e5) {
	      System.out.println(e5.getMessage()); System.exit(1); }
	    if (c1==c2) {
	      s = '*' + s.substring(1);
	    }
	  }

	  GUI.RomKanaPair FoundOne = null;
	  boolean NoneBetter = true;

	  int i,j;
	  j = GUI.Rom2KanaVector.size();
	  GUI.RomKanaPair rk=null;
	  for (i=0;i<j;i++) {
	    try {
	      rk = (GUI.RomKanaPair) GUI.Rom2KanaVector.elementAt(i);
	      if (rk.mode == mode) {
		if (s.startsWith(rk.romaji)) FoundOne = rk; else
		  if (rk.romaji.startsWith(s)) NoneBetter = false;
	      } // else System.out.println(rk.mode + ", " + mode);
	    } catch (Exception e3) { 
	      System.out.println(e3.getMessage()); System.exit(1); }
	  }
	  if ((FoundOne != null) && (NoneBetter)) {
	    Display.setText(LastDisplay+FoundOne.kana);
	    Entry.setText(LastEntry.substring(FoundOne.romaji.length()));
	    Entry.setCaretPosition(Entry.getText().length());
	    LastEntry = Entry.getText();
	    LastDisplay = Display.getText();
	  } else break; /* leave the loop when we can't do more */
	}
      }
    }
    LastEntry = Entry.getText();
    LastDisplay = Display.getText();
  }
  public void keyReleased(KeyEvent e) {
    flushEntry();
  }
}

public class GUI extends Panel implements ActionListener {
  public static class RomKanaPair {
    String romaji;
    String kana;
    int mode;
  }

  LookupAction lua;
  Button fileInputButton;
  JavaDict javadict;
  //  public Checkbox romajiFilterBox;
  public final static String EModeLabel = "R"; /*omaji";*/
  public static String HModeLabel;
  public static String KModeLabel;
  private final static byte[] HModeEUC = {(byte)0244,(byte)0322};
  /*,
    (byte)0244,(byte)0351,
    (byte)0244,(byte)0254,
    (byte)0244,(byte)0312};*/
  private final static byte[] KModeEUC = {(byte)0245,(byte)0253};
  /*
    (byte)0245,(byte)0277,
    (byte)0245,(byte)0253,
    (byte)0245,(byte)0312};*/
  private final static String SearchDirections = "Search Key:";

  private final static String Rom2KanaFile = "romkana.cnv";
  /*  public static ByteToCharConverter toUnicode;
      public static CharToByteConverter fromUnicode;*/
  private final static int MAXLINE = 256;
  public static java.util.Vector Rom2KanaVector;
  static Font dFont = null;
  static Font BigFont = null;

  static {  /* initializer to load Rom2KanaFile into memory */
    Rom2KanaVector = new java.util.Vector(1000);
    /*    try {
	  toUnicode = ByteToCharConverter.getConverter("EUCJIS");
	  fromUnicode = CharToByteConverter.getConverter("EUCJIS");
	  } catch (Exception e) {
	  System.out.println(e.getMessage());
	  System.exit(1);
	  }*/
    
    BufferedReader in = null;
    try {
      File f = new File(Rom2KanaFile);
      FileInputStream tempin = new FileInputStream(f);
      InputStreamReader tempin2 = new InputStreamReader(tempin,"EUCJIS");
      in = new BufferedReader(tempin2);
    } catch (Exception e) {
      System.out.println("While trying to load " + Rom2KanaFile +
		       ": " + e.getMessage());
      System.exit(1);
    }
    try {
      int LineNum = 1;
      int mode = -1; //  -1 = undefined, 0 = hiragana, 1 = katakana
      while (true) {
	// read in a line
	String line;
	line = in.readLine();
	if (line==null) break;
	LineNum++;
	// We have a good line
	if (line.charAt(0)=='#') continue; // continue if comment
	else 
	  if (line.charAt(0)=='$') { /* mode change symbol */
	    switch (line.charAt(1)) {
	    case 'H':
	    case 'h': {
	      mode=0;
	      continue;
	    }
	    case 'K':
	    case 'k': {
	      mode=1;
	      continue;
	    }
	    default: {
	      System.out.println("Syntax error in file: " + Rom2KanaFile +
				 "Line #" + LineNum);
	      System.exit(1);
	    }
	    } /* switch */
	  } else {  // the line should be parsed for rom-kana pairs
	    GUI.RomKanaPair rk = new GUI.RomKanaPair();
	    java.util.StringTokenizer stok = new 
	      java.util.StringTokenizer(line);
	    if (stok.countTokens()<2) {
	      System.out.println("Syntax error in file: " + Rom2KanaFile +
				 "Line #" + LineNum);
	      System.exit(1);
	    } else {
	      try {
		rk.kana = stok.nextToken();
		rk.romaji = stok.nextToken();
		rk.mode = mode;
		if (mode==-1) {
		  System.out.println("Expected $HIRAGANA or $KATAKANA" +
				     " in file: " + Rom2KanaFile +
				     "Line #" + LineNum);
		  System.exit(1);
		}
	      } catch (Exception e) {
		System.out.println(e.getMessage());
		System.exit(1);
	      }
	      Rom2KanaVector.addElement(rk);
	    }
	  }
      }
    } catch (Exception e) { }
  


    // static initializer to convert Labels from EUC
    try {
      HModeLabel = new String(HModeEUC,"EUCJIS");
	//String(toUnicode.convertAll(HModeEUC));
      KModeLabel = new String(KModeEUC,"EUCJIS");
	//String(toUnicode.convertAll(KModeEUC));
    } catch (Exception e) {
      System.out.println(e.getMessage());
      System.exit(1);
    }      
  }
  /*  static String Rom2Kana(String Romaji) {
      
      return " ";
      }*/

  public String dict;
  public String dictindex;
  public GenList.List OutputArea;
  
  static private class LineSelectListener implements ItemListener {
    
    List l;
    JavaDict javadict;
    Button searchbut = null;
    ButtonList bl = null;
    Frame d = null;
    
    public LineSelectListener(List l, JavaDict javadict) {
      this.l=l;
      this.javadict=javadict;
    }

    /*      String s2;
	    s2 = " U"+Integer.toHexString(s.charAt(0))+" ";
	    System.out.println(s2);
	    */

    public void itemStateChanged(ItemEvent e) {
      int i;
      i = l.getSelectedIndex();
      if (i==-1) return;
      l.deselect(i);
      String s;
      s = l.getItem(i);
      if (s==null) return;

      int count=0;
      for (i=0;i<s.length();i++) {
	if (s.charAt(i)>'\u30ff') {
	  count++;
	}
      }
      if (count==0) return;
      char[] carray = new char[count];
      count=0;
      for (i=0;i<s.length();i++) {
	if (s.charAt(i)>'\u30ff') {
	  carray[count]=s.charAt(i);
	  count++;
	}
      }
      // so now carray contains the kanji we want to search for
      javadict.tokanjiviewer.doaction();
      javadict.kanjiview.ViewArray(carray);
    }
  }

  public GUI(JavaDict javadict, Button MenuButton,
	     String dict, String dictindex) {
    this.javadict = javadict;
    this.dict = dict;
    this.dictindex = dictindex;
    Panel f = this;
    f.setLayout(new BorderLayout());
    this.dFont = dFont;

    BigFont = javadict.BigFont;

    //   Output area 
    OutputArea = new GenList.List(5,false);
    OutputArea.setFont(BigFont);
    OutputArea.addItemListener(new LineSelectListener(OutputArea,javadict));

    f.add(OutputArea,"Center");

    //   Input Area
    Panel InputPanel = new Panel();
    Panel InputPanelL = new Panel();
    InputPanel.setLayout(new BorderLayout());

    Choice InputModeChoice = new Choice();
    InputModeChoice.setFont(BigFont);
    
    InputModeChoice.addItem(EModeLabel);


    InputModeChoice.addItem(HModeLabel);
    InputModeChoice.addItem(KModeLabel);
    InputModeChoice.select(EModeLabel);
    InputPanelL.add(InputModeChoice);

    Label DirectionLabel = new Label(SearchDirections);
    InputPanelL.add(DirectionLabel);
    
    TextField EntryField = new TextField(5);
    InputPanelL.add(EntryField);

    Label FeedbackLabel = new Label("                    ");
    FeedbackLabel.setFont(BigFont);
    InputPanel.add(FeedbackLabel,"Center");
    InputPanel.add(InputPanelL,"West");

    /* listener for entry fields */
    EntryField.addKeyListener(new RKDisplay(InputModeChoice,
					    EntryField,
					    FeedbackLabel));
    lua = new LookupAction(InputModeChoice,
			   EntryField,
			   FeedbackLabel,
			   javadict,
			   OutputArea,
			   dict,
			   dictindex);
    EntryField.addActionListener(lua);
    
    Panel ButtonPanel = new Panel(new GridLayout(2,1));
    InputPanel.add(ButtonPanel,"East");



    //    romajiFilterBox = new Checkbox("Romaji",false);
    //    ButtonPanel.add(romajiFilterBox);
    

    fileInputButton = new Button(dict);
    ButtonPanel.add(fileInputButton);
    fileInputButton.addActionListener(this);

    ButtonPanel.add(MenuButton);

    //    InputPanel.add(MenuButton,"East");

    f.add(InputPanel,"South");

  }
  public void actionPerformed(ActionEvent e) {
    FileDialog fileInput = new FileDialog(javadict.frame,
					  "Select a dictionary file",
					  FileDialog.LOAD);
    fileInput.setFile(dict);
    //    fileInput.setFilenameFilter(...);
    fileInput.show();
    String dict2;
    dict2 = fileInput.getFile();
    if ((dict2!=null)&&(dict2.length()>0)) {
      dict=dict2;
      fileInputButton.setLabel(dict);
      lua.setDict(dict);
    }
  }
}
