/*
  Java Japanese-English dictionary/Kanjidic browser
  Copyright (C) 1997 Todd David Rudick

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;

public class HighLightMap extends Canvas implements 
MouseListener,MouseMotionListener,ActionListener,KeyListener {
  Image low;
  Image high;
  boolean on=false;
  boolean forcePaint=true;
  int ox=-99999;
  int oy=-99999;
  int x,y;
  Vector areas = new Vector();
  Vector listeners = new Vector();
  Vector commands = new Vector();
  ActionListener lastAction;
  String lastCommand;
  int w,h;
  Rectangle lastRect = null;

  public void keyReleased(KeyEvent e) { }
  public void keyTyped(KeyEvent e) { }

  public void keyPressed(KeyEvent e) {
    if (e.getKeyCode()==KeyEvent.VK_TAB) {
      Rectangle r = findArea(x,y);
      if (r==null) {
	if (areas.size()>0) {
	  ox=x;oy=y;
	  x=((Rectangle)areas.elementAt(0)).x;
	  y=((Rectangle)areas.elementAt(0)).y;
	  forcePaint=false;
	  paint(getGraphics());
	}
      } else {
	int i;
	i = areas.indexOf(r);
	if (i<areas.size()-1) {
	  ox=x;oy=y;
	  x = ((Rectangle) areas.elementAt(i+1)).x;
	  y = ((Rectangle) areas.elementAt(i+1)).y;
	  forcePaint=false;
	  paint(getGraphics());
	} else if (areas.size()>1) {
	  ox=x;oy=y;
	  x=((Rectangle)areas.elementAt(0)).x;
	  y=((Rectangle)areas.elementAt(0)).y;
	  forcePaint=false;
	  paint(getGraphics());
	}
      }
    } else if (e.getKeyCode()==KeyEvent.VK_SPACE) {
      if (findArea(x,y)!=null) {
	lastAction.actionPerformed(new 
				   ActionEvent(this,
					       ActionEvent.ACTION_PERFORMED,
					       lastCommand));
      }
    } else {

    }
  }

  public void actionPerformed(ActionEvent a) {
    System.out.println("ButtonPress text=" + a.getActionCommand());
  }

  public void addActionListener(ActionListener a, Rectangle r,
				String command) {
    areas.addElement(r);
    listeners.addElement(a);
    commands.addElement(command);
  }

  /*  public Dimension getMinimumSize() {
    return(getPreferredSize());
  }
  public Dimension getMaximumSize() {
    return(getPreferredSize());
  }
  public Dimension getPreferredSize() {
    return(new Dimension(w,h));
  }*/

  Rectangle findArea(int x, int y) {
    Enumeration e = areas.elements();
    Enumeration a = listeners.elements();
    Enumeration c = commands.elements();
    lastAction=null;
    lastCommand=null;
    while (e.hasMoreElements()) {
      Rectangle r;
      r = (Rectangle)e.nextElement();
      ActionListener a2;
      a2 = (ActionListener)a.nextElement();
      String s;
      s = (String)c.nextElement();
      if ((x>=r.x)&&(x<=r.x+r.width)&&
	  (y>=r.y)&&(y<=r.y+r.height)) {
	lastAction=a2;
	lastCommand=s;
	//	System.out.println("Returning  " + lastCommand);
	return(r);
      }
    }
    return(null);
  }

  public void mouseClicked(MouseEvent e) {
    /*    ox=x;oy=y;
	  x=e.getX();
	  y=e.getY();
	  forcePaint=false;
	  paint(getGraphics());
	  lastRect = findArea(x,y);
	  if (lastRect!=null) {
	  lastAction.actionPerformed(new ActionEvent(this,
	  ActionEvent.ACTION_PERFORMED,
	  lastCommand));
	  }*/
  }

  public void mouseEntered(MouseEvent e) {
    ox=x;oy=y;
    x=e.getX();
    y=e.getY();
    forcePaint=false;
    paint(getGraphics());
  }

  public void mouseExited(MouseEvent e) {
    ox=x;oy=y;
    x=-99999;
    y=-99999;
    forcePaint=false;
    paint(getGraphics());
  }

  public void mousePressed(MouseEvent e) {
    ox=x;oy=y;
    x=e.getX();
    y=e.getY();
    forcePaint=false;
    paint(getGraphics());
    lastRect = findArea(x,y);
  }

  public void mouseReleased(MouseEvent e) {
    ox=x;oy=y;
    x=e.getX();
    y=e.getY();
    forcePaint=false;
    paint(getGraphics());
    Rectangle curRect = findArea(x,y);
    
    if ((curRect!=null)&&(lastRect!=null)&&(curRect.equals(lastRect))) {
      lastAction.actionPerformed(new ActionEvent(this,
						 ActionEvent.ACTION_PERFORMED,
						 lastCommand));
    }
  }

  public void mouseDragged(MouseEvent e) {
    ox=x;oy=y;
    x=e.getX();
    y=e.getY();
    forcePaint=false;
    paint(getGraphics());
  }

  public void mouseMoved(MouseEvent e) {
    ox=x;oy=y;
    x=e.getX();
    y=e.getY();
    forcePaint=false;
    paint(getGraphics());
  }

  //  Assumes that both low and high have width = w, height = h
  public HighLightMap(Image low, Image high, int w, int h) {
    this.low=low;
    this.high=high;
    setSize(w,h);
    addMouseMotionListener(this);
    addMouseListener(this);
    addKeyListener(this);
  }

  public void paint(Graphics g) {
    Rectangle r1,r2;
    /*    if (forcePaint) {
	  g.setColor(Color.lightGray);
	  g.fillRect(0,0,getBounds().width,getBounds().height);
	  }*/
    r1=findArea(ox,oy);
    r2=findArea(x,y);
    if ((r1==r2)&&(!forcePaint)) {
      forcePaint=true;
      return;
    }
    else {
      if (r2==null) {
	g.drawImage(low,0,0,this);
	//	System.out.println("no highlight");
      } else {
	g.drawImage(low,0,0,this);
	g.setClip(r2.x,r2.y,r2.width,r2.height);
	g.drawImage(high,0,0,this);
	g.setClip(0,0,getBounds().width,getBounds().height);
	//	System.out.println("hightlight");
      }
    }
    forcePaint=true;
    Toolkit.getDefaultToolkit().sync();
  }
  
  public void update(Graphics g) {
    forcePaint=true;
    paint(g);
  }

  public static void main(String[] args) {
    Frame f = new Frame("test");
    f.setLayout(new BorderLayout());
    Image i1,i2;
    i1 = Toolkit.getDefaultToolkit().getImage("title1.gif");
    i2 = Toolkit.getDefaultToolkit().getImage("title2.gif");
    HighLightMap m = new HighLightMap(i1,i2,600,200);
    f.add(m,"Center");
    f.pack();
    f.show();
    m.addActionListener(m,new Rectangle(273,10,320,24),"JE");
    m.addActionListener(m,new Rectangle(281,58,134,26),"Skip");
    m.addActionListener(m,new Rectangle(281,86,156,21),"Bushu");
    m.addActionListener(m,new Rectangle(281,113,263,24),"HWR");
    m.addActionListener(m,new Rectangle(275,144,85,24),"Options");
    m.addActionListener(m,new Rectangle(274,171,52,25),"Quit");
    m.addActionListener(m,new Rectangle(126,85,90,13),"Warrenty");
    m.addActionListener(m,new Rectangle(125,117,139,26),"Copying");
  }
}
