/*
  Java Japanese-English dictionary/Kanjidic browser
  Copyright (C) 1997 Todd David Rudick

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

import java.awt.*;
import ButtonList;

public class KanjiButtonList extends ButtonList {
  public interface BushuProvider {
    // (suggestion) return '?' if the bushu is out of range
    public abstract char getBushuChar(int bushu);
  }

  boolean allowrare;

  public boolean showColors = true;           // set this for bushu colors
  public BushuProvider bushuProvider = null;  // set this for bushu labels
  public boolean showStrokes = false;         // set this for stroke labels
  //  You probably don't want more than one of the above set at the same time

  int currentBushu = -1;
  int currentStrokes = -1;

  public void removeAll() {
    super.removeAll();
    currentBushu = -1;
    currentStrokes = -1;
  }

  public KanjiButtonList(int rows) { // rows is just for sizing
    super(rows);
  }

  public void setallowrare(boolean allowrare) {
    this.allowrare=allowrare;
  }

  public KanjiButtonList(int rows, ButtonListListener action,
			 boolean allowrare) {
    super(rows,action);
    this.allowrare=allowrare;
  }

  // here's the new stuff, let's color the button according to it's Bushu
  public void add(String s) {

    // if rare kanji are not allowed, filter them out
    // rare is defined as: does not appear in Halpern's dictionary
    if ((!allowrare)&&(s.indexOf(" H",7)==-1)) return;

    // determine Stroke count
    if (showStrokes) {
      int sIndex;
      sIndex = s.indexOf(" S",7);
      if (sIndex!=-1) {
	int sIndex2;
	sIndex2 = s.indexOf(" ",sIndex+1);
	if (sIndex2!=-1) try {
	  int strokes;
	  strokes = Integer.parseInt(s.substring(sIndex+2,sIndex2));
	  if (strokes!=currentStrokes) {
	    currentStrokes = strokes;
	    addLabel(strokes+"",Color.white, Color.black);	  
	  }
	} catch (Exception e9923) {
	}
      }
    }

    // determine Bushu
    int index;
    index = s.indexOf(" B",7);  /* skip the first 7 characters to avoid the 
				   hex-coding which could include any letter */
    if (index==-1) super.add(s);
    else {
      int index2;
      String s2;
      s2 = s.substring(index+2);
      index2 = s2.indexOf(" ");
      if (index2!=-1) s2 = s2.substring(0,index2);
      s2 = s2.trim();
      int bushu;
      try {
	bushu = Integer.parseInt(s2);
	if (bushuProvider!=null) { // add labels
	  if (currentBushu!=bushu) {
	    currentBushu=bushu;
	    char bushC;
	    bushC = bushuProvider.getBushuChar(bushu);
	    addLabel(bushC+"",Color.white, Color.black);
	  }
	}
	if (showColors) {
	  Color fg, bg; 
	  // there are 214 Bushu
	  // we'll need unique colors for each
	  
	  // first, scramble the Bushu so adjacent ones get very 
	  // different colors
	
	  bushu = (bushu*101)%216; // 216 = 6*6*6 (see below)

	  // 6 levels each of red, green, and blue will do it
	  int r,g,b;
	  r = bushu%6;
	  g = (bushu/6)%6;
	  b = (bushu/36)%6;
	  float hsbvals[] = Color.RGBtoHSB(r*42,g*42,b*42,null);
	
	  // keep the brightness away from that nasty area that's so much
	  // trouble to match the fg to
	  if (hsbvals[2]<0.8) hsbvals[2]=(float)(hsbvals[2]*0.5/0.8); 
	  // 0.0-0.5
	  else 
	    hsbvals[2]=(float)((hsbvals[2]/2.0)+0.5); // 0.8-1.0
	
	  bg = Color.getHSBColor(hsbvals[0],hsbvals[1]/2,hsbvals[2]);
	  if (hsbvals[2]<0.80) fg = Color.white;
	  else fg = Color.black;
	  super.add(s,fg,bg);
	} else super.add(s);
      } catch (NumberFormatException e) {
	System.out.println("Kanji encountered with no defined Bushu:"+s2);
	super.add(s);
      }
    }
  }
}
